@@ -1,9 +1,11 @@
 from .dataset_gen import DatasetGenerators
 from .functions import Functions
+from .report_functions import ReportFunctions
 from .generators import Generators
-from .loaders import SpikesFile, Loaders
+from .loaders import Loaders
+from .objects import SpikesFile
 from .plots import Plots
 from .savers import Savers
 from .splitters import Splitters
 from .utils import Utils
-from .main_settings import MainSettings
\ No newline at end of file
+from .main_settings import MainSettings, LocalizationSettings
\ No newline at end of file
