@@ -19,93 +19,171 @@
 ##                                                                             ##
 #################################################################################
 
-import math
-import struct
-import time
 import copy
 import random
-import datetime
+import time
 
-import matplotlib.backends.backend_pdf
-import matplotlib.pyplot as plt
 import numpy as np
 
-from .loaders import SpikesFile
+# from .objects import LocalizationFile
+from .objects import SpikesFile
 from .savers import Savers
 from .utils import Utils
-from .plots import Plots
+
 
 class Functions:
 
 	@staticmethod
 	def check_SpikesFile(spikes_file, settings):
 		"""
-		Checks if the spiking information contained in the SpikesFile is correct and prints "The loaded SpikesFile file has been checked and it's OK" if the file passes all the checks.
+        Checks if the spiking information contained in the SpikesFile is correct and prints "The loaded SpikesFile file has been checked and it's OK" if the file passes all the checks.
+
+        Parameters:
+                spikes_file (SpikesFile): File to check.
+                settings (MainSettings): Configuration parameters for the file to check.
+
+        Returns:
+                None.
+
+        Raises:
+                TimestampOrderError: If the SpikesFile contains at least one timestamp which value is less than 0.
+                TimestampOrderError: If the SpikesFile contains at least one timestamp that is lesser than its previous one.
+                AddressValueError: If the SpikesFile contains at least one address less than 0 or greater than the num_channels that you specified in the MainSettings.
+        Notes:
+                If mono_stereo is set to 1 (stereo) in the MainSettings, then  addresses should be less than num_channels*2.
+
+                If on_off_both is set to 1 (both) in the MainSettings, then addresses should be less than num_channels*2.
+
+                If mono_stereo is set to 1 and on_off_both is set to 1 in the MainSettings, then addresses should be less than num_channels*2*2.
+        """
+		# Convert to numpy arrays
+		addresses = np.array(spikes_file.addresses, copy=False)
+		timestamps = np.array(spikes_file.timestamps, copy=False)
+
+		# Check if all timestamps are greater than zero
+		any_negative = np.any(timestamps < 0)
+
+		if any_negative:
+			print("[Functions.check_SpikesFile] > TimestampOrderError: The SpikesFile file that you loaded has at least one timestamp that is less than 0.")
+
+		# Check if each timestamp is greater than its previous one
+		increasing_order = np.all(timestamps[:-1] <= timestamps[1:])
+
+		if not increasing_order:
+			print("[Functions.check_SpikesFile] > TimestampOrderError: The SpikesFile file that you loaded has at least one timestamp whose value is lesser than its previous one.")
+
+		# Calculate maximum number of addresses
+		number_of_addresses = settings.num_channels * (settings.on_off_both + 1) * (settings.mono_stereo + 1)
+
+		# Check if all addresses are between zero and the total number of addresses
+		all_in_range = np.all((addresses >= 0) & (addresses < number_of_addresses))
+
+		if not all_in_range:
+			print("[Functions.check_SpikesFile] > AddressValueError: The SpikesFile file that you loaded has at least one event whose address is either less than 0 or greater than the number of addresses that you specified.")
+
+		# Check if all is OK
+		all_ok = not any_negative and increasing_order and all_in_range
+
+		if all_ok:
+			print("[Functions.check_SpikesFile] > The loaded SpikesFile file has been checked and it's OK")
+
+		return not any_negative, increasing_order, all_in_range
+
+	@staticmethod
+	def check_LocalizationFile(localization_file, settings, localization_settings):
+		"""
+		Checks if the spiking information contained in the LocalizationFile is correct and prints "The loaded LocalizationFile file has been checked and it's OK" if the file passes all the checks.
 		
 		Parameters:
-				spikes_file (SpikesFile): File to check.
+				localization_file (LocalizationFile): File to check.
 				settings (MainSettings): Configuration parameters for the file to check.
+				localization_settings (LocalizationSettings): Configuration parameters of the localization model for the file to check.
 
 		Returns:
 				None.
 		
 		Raises:
-				TimestampOrderError: If the SpikesFile contains at least one timestamp which value is less than 0.
-				TimestampOrderError: If the SpikesFile contains at least one timestamp that is lesser than its previous one.
-				AddressValueError: If the SpikesFile contains at least one address less than 0 or greater than the num_channels that you specified in the MainSettings.
+				TimestampOrderError: If the LocalizationFile contains at least one timestamp which value is less than 0.
+				TimestampOrderError: If the LocalizationFile contains at least one timestamp that is lesser than its previous one.
+				ChannelValueError: If the LocalizationFile contains at least one address less than mso_start_channel or greater than mso_end_channel that you specified in the LocalizationSettings.
+				NeuronIDValueError: If the LocalizationFile contains at least one address less than 0 or greater than the mso_num_neurons_channel you specified in LocalizationSettings
 		Notes:   
-				If mono_stereo is set to 1 (stereo) in the MainSettings, then  addresses should be less than num_channels*2.
-
-				If on_off_both is set to 1 (both) in the MainSettings, then addresses should be less than num_channels*2.
-				
-				If mono_stereo is set to 1 and on_off_both is set to 1 in the MainSettings, then addresses should be less than num_channels*2*2.
+				If mso_start_channel is set to 33 and mso_end_channel is set to 36, there will be four possible channel values: [33, 36]
 		"""
 
-		if settings.on_off_both == 1:
-			number_of_addresses = settings.num_channels*2
-		else:
-			number_of_addresses = settings.num_channels
 		# Check if all timestamps are greater than zero
-		a = all(item >= 0  for item in spikes_file.timestamps)
+		a = all(item >= 0  for item in localization_file.mso_timestamps)
 
 		if not a:
-			print("[Functions.check_SpikesFile] > TimestampOrderError: The SpikesFile file that you loaded has at least one timestamp that is less than 0.")
+			print("[Functions.check_LocalizationFile] > TimestampOrderError: The LocalizationFile file that you loaded has at least one timestamp that is less than 0.")
 
 		# Check if each timestamp is greater than its previous one
-		b = not any(i > 0 and spikes_file.timestamps[i] < spikes_file.timestamps[i-1] for i in range(len(spikes_file.timestamps)))
+		b = not any(i > 0 and localization_file.mso_timestamps[i] < localization_file.mso_timestamps[i-1] for i in range(len(localization_file.mso_timestamps)))
 
 		if not b:
-			print("[Functions.check_SpikesFile] > TimestampOrderError: The SpikesFile file that you loaded has at least one timestamp whose value is lesser than its previous one.")
+			print("[Functions.check_LocalizationFile] > TimestampOrderError: The LocalizationFile file that you loaded has at least one timestamp whose value is lesser than its previous one.")
 
-		# Check if all addresses are between zero and the total number of addresses
-		c = all(item >= 0 and item < number_of_addresses*(settings.mono_stereo + 1) for item in spikes_file.addresses)
+		# Check if all channel values are between mso_start_channel and mso_end_channel
+		c = all(item >= localization_settings.mso_start_channel and item <= localization_settings.mso_end_channel for item in localization_file.mso_channels)
 
 		if not c:
-			print("[Functions.check_SpikesFile] > AddressValueError: The SpikesFile file that you loaded has at least one event whose address is either less than 0 or greater than the number of addresses that you specified.")
+			print("[Functions.check_LocalizationFile] > ChannelValueError: The LocalizationFile file that you loaded has at least one event whose channel value is either less than mso_start_channel or greater than mso_end_channel.")
 
-		if a and b and c:
-			print("[Functions.check_SpikesFile] > The loaded SpikesFile file has been checked and it's OK")
-				
+		# Check if all neuron IDs are between zero and the number of mso_num_neurons_channel
+		d = all(item >= 0 and item < localization_settings.mso_num_neurons_channel for item in localization_file.mso_neuron_ids)
 
-	@staticmethod 
-	def adapt_SpikesFile(spikes_file, settings):
+		if not d:
+			print("[Functions.check_LocalizationFile] > NeuronIDValueError: The LocalizationFile file that you loaded has at least one event whose neuron ID value is either less than zero or greater than mso_num_neurons_channel.")
+
+		if a and b and c and d:
+			print("[Functions.check_LocalizationFile] > The loaded LocalizationFile file has been checked and it's OK")
+
+	@staticmethod
+	def adapt_timestamps(spikes_file, settings):
 		"""
-		Subtracts the smallest timestamp of the SpikesFile to all of the timestamps contained in the file (in order to start from 0)
+		Subtracts the smallest timestamp of the timestamps list to all of the timestamps contained in the list (in order to start from 0)
 		It also adapts timestamps based on the tick frequency (ts_tick in the MainSettings).
-		
+
 		Parameters:
-				spikes_file (SpikesFile): File to adapt.
+		  		spikes_file:
 				settings (MainSettings): Configuration parameters for the file to adapt.
 
 		Returns:
-				SpikesFile:  Adapted SpikesFile.
+				adapted_timestamps:  Adapted timestamps list.
 		"""
-		minimum_ts = min(spikes_file.timestamps)
-		if settings.reset_timestamp:
-			spikes_file.timestamps = [(x - minimum_ts)*settings.ts_tick for x in spikes_file.timestamps]
+		if spikes_file.timestamps != []:
+			# Convert to numpy array
+			timestamps = np.array(spikes_file.timestamps, copy=False)
+
+			if settings.reset_timestamp:
+				# Substract the minimum to all timestamps
+				minimum_ts = spikes_file.min_ts
+				adapted_timestamps = (timestamps - minimum_ts) * settings.ts_tick
+
+				# Update the maximum and minimum values
+				spikes_file.max_ts = (spikes_file.max_ts - minimum_ts) * settings.ts_tick
+				spikes_file.min_ts = 0
+			else:
+				adapted_timestamps = timestamps * settings.ts_tick
+
+				# Update the maximum and minimum values
+				spikes_file.max_ts = adapted_timestamps[spikes_file.max_ts_index] * settings.ts_tick
+				spikes_file.min_ts = adapted_timestamps[spikes_file.min_ts_index] * settings.ts_tick
+
+			# Update the timestamps
+			spikes_file.timestamps = adapted_timestamps.astype(dtype=np.dtype(">u" + str(settings.timestamp_size)))
 		else:
-			spikes_file.timestamps = [x*settings.ts_tick for x in spikes_file.timestamps]
-		return spikes_file
+			print("[Functions.adapt_timestamps] > The SpikesFile timestamps are empty")
+
+
+	@staticmethod
+	def order_SpikesFile(spikes_file, settings):
+		# Indices that would sort the file
+		indexes = np.argsort(spikes_file.timestamps)
+
+		# Sort the arrays
+		spikes_file.addresses = spikes_file.addresses[indexes]
+		spikes_file.timestamps = spikes_file.timestamps[indexes]
 
 
 	@staticmethod
@@ -129,7 +207,7 @@ class Functions:
 			prevSpike = [None] * (settings.num_channels) * (1 + settings.mono_stereo)
 			phaseLockedAddrs = []
 			phaseLockedTs = []
-			for i in range(len(spikes_file.addresses)):
+			for i in range(len(spikes_file.addressebrew install openssl@1.1s)):
 				if prevSpike[spikes_file.addresses[i]//2] == None:
 					prevSpike[spikes_file.addresses[i]//2] = spikes_file.addresses[i]%2
 				else:
@@ -141,6 +219,10 @@ class Functions:
 						prevSpike[spikes_file.addresses[i]//2] = spikes_file.addresses[i]%2
 			spikes_file = SpikesFile([], [])
 			spikes_file.addresses = phaseLockedAddrs
+			spikes_file.min_ts = np.min(phaseLockedTs)
+			spikes_file.max_ts = np.max(phaseLockedTs)
+			spikes_file.min_ts_index = np.argmin(phaseLockedTs)
+			spikes_file.max_ts_index = np.argmax(phaseLockedTs)
 			spikes_file.timestamps = phaseLockedTs
 			return spikes_file
 		else:
@@ -175,15 +257,15 @@ class Functions:
 			spikes_file_mono = Utils.extract_addr_and_ts(addr_ts)
 			if left_right:
 				spikes_file_mono.addresses = [x-left_right*settings.num_channels*(settings.on_off_both + 1) for x in spikes_file_mono.addresses]
-			
-			
+
+
 			if return_save_both == 0:
 				return spikes_file_mono
 			elif return_save_both == 1 or return_save_both == 2:
-				Savers.save_as_any(spikes_file_mono, path=path, output_format=output_format, settings=settings) 
+				Savers.save_as_any(spikes_file_mono, path=path, output_format=output_format, settings=settings)
 				if return_save_both == 2:
 					return spikes_file_mono
-			
+
 		else:
 			print("[Functions.stereo_to_mono] > SettingsError: this functionality cannot be performed over a mono aedat file.")
 
@@ -207,7 +289,7 @@ class Functions:
 		Raises:
 				SettingsError: If the input file is a stereo SpikesFile (settings.mono_stereo is set to 1).
 
-		Note:	A positive delay value means that spikes from the right cochlea will be generated after the ones from the left cochlea, and vice versa.
+		Note:	The timestamp of the left event is used as reference. Thus, the timestamp of the right event will be ts_right = ts_left + delay.
 		"""
 
 		if settings.mono_stereo == 0:
@@ -228,13 +310,12 @@ class Functions:
 			elif return_save_both == 1 or return_save_both == 2:
 				#settings_new = copy.deepcopy(settings)
 				#settings_new.mono_stereo = 1
-				Savers.save_as_any(spikes_file_new, path=path, output_format=output_format, settings=settings) 
+				Savers.save_as_any(spikes_file_new, path=path, output_format=output_format, settings=settings)
 				if return_save_both == 2:
 					return spikes_file_new
 
 		else:
-			print("[Functions.mono_to_stereo] > SettingsError: this functionality cannot be performed over a stereo aedat file.")        
-
+			print("[Functions.mono_to_stereo] > SettingsError: this functionality cannot be performed over a stereo aedat file.")
 
 
 	@staticmethod
@@ -265,83 +346,28 @@ class Functions:
 		if reset_addresses == True:
 			new_spikes_file.addresses = [addr - addresses[0] for addr in new_spikes_file.addresses]
 		new_spikes_file.timestamps = spikes_per_channels_ts
-		new_spikes_file = order_timestamps(new_spikes_file)
+		new_spikes_file = Utils.order_timestamps(new_spikes_file)
 		return new_spikes_file
 
 
-	
-
 	@staticmethod
-	def PDF_report(spikes_file, settings, output_path, plots = ["Spikegram", "Sonogram", "Histogram", "Average activity", "Difference between L/R"], vector = False, verbose = False):
+	def ISI(spikes_file, verbose = False):
 		"""
-		Generates a PDF report with the spikegram, sonogram, histogram, average activity and difference between L/R plots obtained from the input SpikesFile or path containing SpikeFiles.
+		Generates an array with the inter-spike intervals of the input SpikesFile.
 		
 		Parameters:
 				spikes_file (SpikesFile or string): File or path to use.
 				settings (MainSettings): Configuration parameters for the input file.
-				vector (boolean, optional): Set to True if you want the Spikegram plot vectorized. Note: this may make your PDF heavy.
 				verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.
 
 		Returns:
-				None.
-		
-		Notes:   
-				If the path used as input is a folder instead of a spikes file, the PDF report is generated for every spikes file contained in the folder.
+				int[]: inter-spike interval array.
 		"""
 
-		if isinstance(spikes_file, str):
-    			
-			spikes_file_extension = os.path.splitext(spikes_file)
+		if verbose == True: start_time = time.time()
 
-			if spikes_file_extension == ".aedat":
-				spikes_file = Loaders.loadAEDAT(spikes_file, settings)
-			elif spikes_file_extension == ".csv":
-				spikes_file = Loaders.loadCSV(spikes_file, settings)
-			elif spikes_file_extension == ".txt":
-				spikes_file = Loaders.loadZynqGrabberData(spikes_file, settings)
-    				
-			spikes_file = Functions.adapt_SpikesFile(spikes_file, settings)
-		
-		if isinstance(spikes_file, SpikesFile):
-    		
-			pdf = matplotlib.backends.backend_pdf.PdfPages(output_path)
-			
-			# Spikegram
-			if any("Spikegram" in s for s in plots):
-				spikegram = Plots.spikegram(spikes_file, settings, )
-				pdf.savefig(spikegram)
-				plt.draw()
-			# Sonogram
-			if any("Sonogram" in s for s in plots):
-				sonogram = Plots.sonogram(spikes_file, settings)		
-				pdf.savefig(sonogram)
-				plt.draw()
-			# Histogram
-			if any("Histogram" in s for s in plots):
-				Plots.histogram(spikes_file, settings,)		
-				pdf.savefig()
-				plt.draw()
-			# Average activity
-			if any("Average activity" in s for s in plots):
-				Plots.average_activity(spikes_file, settings,)		
-				pdf.savefig()
-				plt.draw()
-			# Difference between L/R
-			if settings.mono_stereo == 1 and any("Difference between L/R" in s for s in plots):		
-				Plots.difference_between_LR(spikes_file, settings,)		
-				pdf.savefig()
-				plt.draw()
+		isi_array = np.diff(spikes_file.timestamps)
 
+		if verbose == True: print('ISI CALCULATION', time.time() - start_time)
 
-			d = pdf.infodict()
-			d['Title'] = 'pyNAVIS report'
-			d['Author'] = 'Juan P. Dominguez-Morales'
-			d['Subject'] = 'pyNAVIS report'
-			d['Keywords'] = 'pyNAVIS'
-			d['CreationDate'] = datetime.datetime.today()
-			d['ModDate'] = datetime.datetime.today()
-
-			pdf.close()
-
-		else:
-			print("[Functions.PDF_report] > InputFileError: the input SpikesFile is not valid.")
\ No newline at end of file
+		return isi_array
\ No newline at end of file
